{
    This file is part of the Web Service Toolkit
    Copyright (c) 2006 by Inoussa OUEDRAOGO

    This file is provide under LGPL licence which can be found at
    http://www.gnu.org/licenses/lgpl.html.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
}

unit basestubimp;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Contnrs,
  formaterintf;

Type

  ERemoteException = EBaseException;
  
  { TAbstractRemoteCallMaker }

  TAbstractRemoteCallMaker = class
  Public
    constructor Create();virtual;abstract;
    procedure MakeCall(ASerializer : TAbstractFormatter);virtual;abstract;
  End;

  TAbstractRemoteCallMakerClass = class of TAbstractRemoteCallMaker;
  { TBaseStub }

  TBaseProxy = Class(TInterfacedObject)
  Private
    FTarget : String;
    FAdress : String;
    FSerializerClass : TAbstractFormatterClass;
  Protected
    function GetAdress():String;
    function GetTarget():String;
    function GetProtocol():String;
    function CreateSerializer():TAbstractFormatter;
    procedure MakeCall(ASerializer : TAbstractFormatter);
  Public
    constructor Create(
      Const ATarget, AAdress : String;
            ASerializerClass : TAbstractFormatterClass
    );virtual;
  End;

  { TRemoteCallMakerRegItem }

  TRemoteCallMakerRegItem = class
  private
    FRemoteCallMakerClass: TAbstractRemoteCallMakerClass;
    FName: String;
  Public
    constructor Create(AName : String;ARemoteCallMakerClass : TAbstractRemoteCallMakerClass);
    Property RemoteCallMakerClass : TAbstractRemoteCallMakerClass Read FRemoteCallMakerClass;
    Property Name : String Read FName;
  End;

  { TRemoteCallMakerRegistry }

  TRemoteCallMakerRegistry = class // must be threadsafe????
  Private
    FList : TObjectList;
    function GetCount: Integer;
    function GetItem(Index: Integer): TRemoteCallMakerRegItem;
  Public
    constructor Create();
    destructor Destroy();override;
    function IndexOf(Const AName : String):Integer;
    function Register(Const AName : String;ARemoteCallMakerClass : TAbstractRemoteCallMakerClass):Integer;
    function Find(Const AName : String):TRemoteCallMakerRegItem;
    Property Count : Integer Read GetCount;
    Property Item[Index:Integer] : TRemoteCallMakerRegItem Read GetItem;default;
  End;

  function GetRemoteCallMakerRegistry() : TRemoteCallMakerRegistry;
  
implementation

{ TBaseProxy }

function TBaseProxy.GetAdress(): String;
begin
  Result := FAdress;
end;

function TBaseProxy.GetTarget(): String;
begin
  Result := FTarget;
end;

function TBaseProxy.GetProtocol(): String;
Var
  i : Integer;
begin
  i := AnsiPos(':',FAdress);
  If ( i > 0 ) Then
    Result:= Copy(FAdress,1,Pred(i))
  Else
    Result := 'HTTP';
end;

function TBaseProxy.CreateSerializer(): TAbstractFormatter;
begin
  Result := FSerializerClass.Create();
end;

procedure TBaseProxy.MakeCall(ASerializer: TAbstractFormatter);
Var
  i : Integer;
  rmtCallMkr : TAbstractRemoteCallMaker;
begin
  i := GetRemoteCallMakerRegistry().IndexOf(GetProtocol());
  If ( i < 0 ) Then
    Raise ERemoteException.Create('No "RemoteCallMaker" registered.');
  rmtCallMkr := GetRemoteCallMakerRegistry()[i].RemoteCallMakerClass.Create();
  Try
    rmtCallMkr.MakeCall(ASerializer);
  Finally
    rmtCallMkr.Free();
  End;
end;

constructor TBaseProxy.Create(
  Const ATarget, AAdress : String;
        ASerializerClass : TAbstractFormatterClass
);
begin
  Assert(Assigned(ASerializerClass));
  FTarget := ATarget;
  FAdress := AAdress;
  FSerializerClass := ASerializerClass;
end;


{ TRemoteCallMakerRegistry }

function TRemoteCallMakerRegistry.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TRemoteCallMakerRegistry.GetItem(Index: Integer): TRemoteCallMakerRegItem;
begin
  Result := FList[Index] as TRemoteCallMakerRegItem;
end;

constructor TRemoteCallMakerRegistry.Create();
begin
  FList := TObjectList.Create(True);
end;

destructor TRemoteCallMakerRegistry.Destroy();
begin
  FreeAndNil(FList);
  inherited Destroy();
end;

function TRemoteCallMakerRegistry.IndexOf(const AName: String): Integer;
Var
  c : Integer;
begin
  c := Pred(Count);
  For Result := 0 To c Do Begin
    If AnsiSameText(AName,Item[Result].Name) Then
      Exit;
  End;
  Result := -1;
end;

function TRemoteCallMakerRegistry.Register(
  Const AName : String;
        ARemoteCallMakerClass : TAbstractRemoteCallMakerClass
):Integer;
begin
  Result := IndexOf(AName);
  If ( Result = -1 ) Then
    Result := FList.Add(TRemoteCallMakerRegItem.Create(AName,ARemoteCallMakerClass));
end;

function TRemoteCallMakerRegistry.Find(const AName: String): TRemoteCallMakerRegItem;
Var
  i : Integer;
begin
  i := IndexOf(AName);
  If ( i > -1 ) Then
    Result := Item[i]
  Else
    Result := Nil;
end;

{ TRemoteCallMakerRegItem }

constructor TRemoteCallMakerRegItem.Create(
  AName: String;
  ARemoteCallMakerClass: TAbstractRemoteCallMakerClass
);
begin
  Assert(Assigned(ARemoteCallMakerClass));
  FName := AName;
  FRemoteCallMakerClass := ARemoteCallMakerClass;
end;

Var
  RemoteCallMakerRegistryInst : TRemoteCallMakerRegistry = Nil;
  
function GetRemoteCallMakerRegistry() : TRemoteCallMakerRegistry;
Begin
  Result := RemoteCallMakerRegistryInst;
End;

Initialization
  RemoteCallMakerRegistryInst := TRemoteCallMakerRegistry.Create();
  
Finalization
  FreeAndNil(RemoteCallMakerRegistryInst);
  
end.

