{
    This file is part of the Web Service Toolkit
    Copyright (c) 2006 by Inoussa OUEDRAOGO

    This file is provide under LGPL licence which can be found at
    http://www.gnu.org/licenses/lgpl.html.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
}

unit formaterintf;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, TypInfo;

Type
  EBaseException = class(Exception)
  End;

  EFormatterException = class(EBaseException)
  End;

  { TAbstractFormatter }

  TAbstractFormatterClass = class of TAbstractFormatter;
  TAbstractFormatter = class
  Private
    FCallProcedureName : string;
    FCallTarget:string;
    FCallAdress:string;
  Protected
    function GetCurrentScope():String;virtual;abstract;
    procedure Error(Const AMsg : String);overload;
    procedure Error(Const AMsg : String;Const AArgs : Array Of Const);overload;
  public
    constructor Create();virtual;abstract;
    procedure Clear();virtual;
    procedure BeginScope(
      Const AScopeName,ANameSpace : string;
      Const ANameSpaceShortName   : string = ''
    );virtual;abstract;
    procedure EndScope();virtual;abstract;
    procedure AddScopeAttribute(Const AName,AValue : string);virtual;abstract;
    Procedure BeginScopeRead(Const AScopeName: string);virtual;abstract;
    procedure EndScopeRead();virtual;abstract;

    property CurrentScope : String Read GetCurrentScope;
    
    procedure Put(
      Const AName     : String;
      Const ATypeInfo : PTypeInfo;
      Const AData
    );virtual;abstract;
    procedure Get(
      Const ATypeInfo : PTypeInfo;
      Var   AName     : String;
      Var   AData
    );virtual;abstract;

    procedure BeginCall(Const AProcName,ATarget,ACallAdress:string);virtual;
    procedure EndCall();virtual;abstract;
    procedure BeginCallRead();virtual;abstract;
    function GetCallProcedureName():String;
    function GetCallTarget():String;
    function GetCallAdress():String;
    
    procedure SaveToStream(AStream : TStream);virtual;abstract;
    procedure LoadFromStream(AStream : TStream);virtual;abstract;
  End;


implementation

{ TAbstractFormatter }

procedure TAbstractFormatter.Error(const AMsg: String);
begin
  Raise EFormatterException.Create(AMsg);
end;

procedure TAbstractFormatter.Error(const AMsg: String;const AArgs: array of const);
begin
  Raise EFormatterException.CreateFmt(AMsg,AArgs);
end;

procedure TAbstractFormatter.Clear();
begin
end;

procedure TAbstractFormatter.BeginCall(Const AProcName,ATarget,ACallAdress:string);
begin
  FCallProcedureName := AProcName;
  FCallTarget := ATarget;
  FCallAdress := ACallAdress;
end;

function TAbstractFormatter.GetCallProcedureName(): String;
begin
  Result := FCallProcedureName;
end;

function TAbstractFormatter.GetCallTarget(): String;
begin
  Result := FCallTarget;
end;

function TAbstractFormatter.GetCallAdress(): String;
begin
  Result := FCallAdress;
end;


end.
