{
    This file is part of the Web Service Toolkit
    Copyright (c) 2006 by Inoussa OUEDRAOGO

    This file is provide under LGPL licence which can be found at
    http://www.gnu.org/licenses/lgpl.html.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
}
unit xml_formater;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, TypInfo, Contnrs, DOM,
  formaterintf, baseobject;
  
Const
  sXML_NS = 'xmlns';
  sXSI_NS = 'http://www.w3.org/1999/XMLSchema-instance';
  sXSI_TYPE = 'xsi:type';
  sSOAP_ENV = 'http://schemas.xmlsoap.org/soap/envelope/';
  sSOAP_ENV_ABR = 'SOAP-ENV';
  sXSD_NS = 'http://www.w3.org/1999/XMLSchema';
  sXSD = 'xsd';

Type

  { TXmlFormatter }

  TXmlFormatterEncodingStyle = ( xfesEncoded, xfesLitteral );

  TXmlFormatterClass = class of TXmlFormatter;
  TXmlFormatter = class(TAbstractFormatter)
  private
    FEncodingStyle: TXmlFormatterEncodingStyle;
    FNameSpaceCounter : Integer;
    FDoc : TXMLDocument;
    FStack : TObjectStack;

    procedure ClearStack();

    function NextNameSpaceCounter():Integer;
    function HasScope():Boolean;
    function FindAttributeByValueInScope(Const AAttValue : String):String;

    procedure CheckScope();
    function InternalPutData(
      Const AName     : String;
      Const ATypeInfo : PTypeInfo;
      Const AData     : string
    ):TDOMNode;
    function PutBool(
      Const AName     : String;
      Const ATypeInfo : PTypeInfo;
      Const AData     : Boolean
    ):TDOMNode;
    function PutInt(
      Const AName     : String;
      Const ATypeInfo : PTypeInfo;
      Const AData     : Integer
    ):TDOMNode;
    function PutStr(
      Const AName     : String;
      Const ATypeInfo : PTypeInfo;
      Const AData     : String
    ):TDOMNode;
    procedure PutObj(
      Const AName     : String;
      Const ATypeInfo : PTypeInfo;
      Const AData     : TObject
    );

    function GetNodeValue(Const AName : String):DOMString;
    procedure GetBool(
      Const ATypeInfo : PTypeInfo;
      Var   AName     : String;
      Var   AData     : Boolean
    );
    procedure GetInt(
      Const ATypeInfo : PTypeInfo;
      Var   AName     : String;
      Var   AData     : Integer
    );
    procedure GetStr(
      Const ATypeInfo : PTypeInfo;
      Var   AName     : String;
      Var   AData     : String
    );
    procedure GetObj(
      Const ATypeInfo : PTypeInfo;
      Var   AName     : String;
      Var   AData     : TObject
    );
  protected
    function GetCurrentScope():String;override;
    function GetCurrentScopeObject():TDOMElement;
  public
    constructor Create();override;
    destructor Destroy();override;
    procedure Clear();override;

    procedure BeginScope(
      Const AScopeName,ANameSpace : string;
      Const ANameSpaceShortName   : string = ''
    );override;
    procedure EndScope();override;
    procedure AddScopeAttribute(Const AName,AValue : string);override;
    Procedure BeginScopeRead(Const AScopeName: string);override;
    procedure EndScopeRead();override;

    procedure Put(
      Const AName     : String;
      Const ATypeInfo : PTypeInfo;
      Const AData
    );override;
    procedure Get(
      Const ATypeInfo : PTypeInfo;
      Var   AName     : String;
      Var   AData
    );override;

    procedure BeginCall(Const AProcName,ATarget,ACallAdress:string);override;
    procedure EndCall();override;
    procedure BeginCallRead();override;

    procedure SaveToStream(AStream : TStream);override;
    procedure LoadFromStream(AStream : TStream);override;

    Property EncodingStyle : TXmlFormatterEncodingStyle Read FEncodingStyle Write FEncodingStyle;
  End;

  { TSOAPRemotableComplexObject }

  TSOAPRemotableComplexObject = class(TBaseRemotable)
  Public
    class procedure Save(
            AObject   : TBaseRemotable;
            AStore    : TAbstractFormatter;
      Const AName     : String;
      Const ATypeInfo : PTypeInfo
    );override;
    class procedure Load(
            AObject   : TBaseRemotable;
            AStore    : TAbstractFormatter;
      var   AName     : String;
      const ATypeInfo : PTypeInfo
    );override;
  End;


  function CreateXmlFormater():TAbstractFormatter;
  function GetSoapXmlFormaterClass():TAbstractFormatterClass;


implementation
Uses XMLWrite, XMLRead, type_registry;

function IsStrEmpty(Const AStr:String):Boolean;
begin
  Result := ( Length(Trim(AStr)) = 0 );
end;

function GetSoapXmlFormaterClass():TAbstractFormatterClass;
begin
  Result := TXmlFormatter;
end;

function CreateXmlFormater():TAbstractFormatter;
begin
  Result := TXmlFormatter.Create();
end;

{ TXmlFormatter }

procedure TXmlFormatter.ClearStack();
Var
  i, c : Integer;
begin
  c := FStack.Count;
  For I := 1 To c Do
    FStack.Pop();
end;

procedure TXmlFormatter.BeginScopeRead(Const AScopeName: string);
Var
  locNode : TDOMNode;
begin
  locNode := GetCurrentScopeObject().FindNode(AScopeName);
  If Not Assigned(locNode) Then
    Error('Scope not found : "%s"',[AScopeName]);
  FStack.Push(locNode);
end;

procedure TXmlFormatter.EndScopeRead();
begin
  FStack.Pop();
end;

function TXmlFormatter.NextNameSpaceCounter(): Integer;
begin
  Inc(FNameSpaceCounter);
  Result := FNameSpaceCounter;
end;

function TXmlFormatter.HasScope(): Boolean;
begin
  Result := Assigned(FStack.Peek);
end;

function TXmlFormatter.FindAttributeByValueInScope(const AAttValue: String): String;
Var
  i,c : Integer;
  tmpNode : TDOMNode;
begin
  If HasScope() Then Begin
    tmpNode := GetCurrentScopeObject();
    While Assigned(tmpNode) Do Begin
      If Assigned(tmpNode.Attributes) Then Begin
        c := Pred(tmpNode.Attributes.Length);
        For i := 0 To c Do Begin
          If AnsiSameText(AAttValue,tmpNode.Attributes.Item[i].NodeValue) Then Begin
            Result := tmpNode.Attributes.Item[i].NodeName;
            Exit;
          End;
        End;
      End;
      tmpNode := tmpNode.ParentNode;
    End;
  End;
  Result := '';
end;

procedure TXmlFormatter.CheckScope();
begin
  If Not HasScope() Then
    Error('There is no scope.');
end;

function TXmlFormatter.InternalPutData(
  const AName      : String;
  const ATypeInfo  : PTypeInfo;
  const AData      : string
): TDOMNode;
Var
  namespaceLongName, namespaceShortName : string;
  regItem : TTypeRegistryItem;
begin
  Result := FDoc.CreateElement(AName);
  Result.AppendChild(FDoc.CreateTextNode(AData));
  GetCurrentScopeObject().AppendChild(Result);
  If ( EncodingStyle = xfesEncoded ) Then Begin
    regItem := GetTypeRegistry().ItemByTypeInfo[ATypeInfo];
    namespaceLongName := regItem.NameSpace;
    namespaceShortName := FindAttributeByValueInScope(namespaceLongName);
    If IsStrEmpty(namespaceShortName) Then Begin
      namespaceShortName := Format('ns%d',[NextNameSpaceCounter()]);
      AddScopeAttribute(sXML_NS + ':'+namespaceShortName,namespaceLongName);
    End Else Begin
      namespaceShortName := Copy(namespaceShortName,AnsiPos(':',namespaceShortName) + 1,MaxInt);
    End;
    (Result As TDOMElement).SetAttribute(sXSI_TYPE,Format('%s:%s',[namespaceShortName,regItem.DeclaredName]));
  End;
end;

function TXmlFormatter.PutBool(
  const AName     : String;
  const ATypeInfo : PTypeInfo;
  const AData     : Boolean
): TDOMNode;
begin
  Result := InternalPutData(AName,ATypeInfo,LowerCase(BoolToStr(AData)));
end;

function TXmlFormatter.PutInt(
  const AName: String;
  const ATypeInfo: PTypeInfo;
  const AData: Integer
):TDOMNode;
begin
  Result := InternalPutData(AName,ATypeInfo,IntToStr(AData));
end;

function TXmlFormatter.PutStr(
  const AName: String;
  const ATypeInfo: PTypeInfo;
  const AData: String
):TDOMNode;
begin
  Result := InternalPutData(AName,ATypeInfo,AData);
end;

procedure TXmlFormatter.PutObj(
  const AName      : String;
  const ATypeInfo  : PTypeInfo;
  const AData      : TObject
);
begin
  TBaseRemotableClass(GetTypeData(ATypeInfo)^.ClassType).Save(AData As TBaseRemotable, Self,AName,ATypeInfo);
end;

function TXmlFormatter.GetNodeValue(const AName: String): DOMString;
Var
  locElt : TDOMElement;
begin
  locElt := GetCurrentScopeObject().FindNode(AName) As TDOMElement;
  If Assigned(locElt) Then Begin
    If locElt.HasChildNodes Then
      Result := locElt.FirstChild.NodeValue
    Else
      Result := locElt.NodeValue;
  End Else Begin
    Error('Param not found : "%s"',[AName]);
  End;
end;

procedure TXmlFormatter.GetBool(
  const ATypeInfo  : PTypeInfo;
  var   AName      : String;
  var   AData      : Boolean
);
Var
  locBuffer : String;
begin
  locBuffer := LowerCase(Trim(GetNodeValue(AName)));
  If IsStrEmpty(locBuffer) Then
    AData := False
  Else
    AData := StrToBool(locBuffer);
end;

procedure TXmlFormatter.GetInt(
  const ATypeInfo: PTypeInfo;
  var   AName: String;
  var   AData: Integer
);
Var
  locBuffer : String;
begin
  locBuffer := Trim(GetNodeValue(AName));
  If IsStrEmpty(locBuffer) Then
    AData := 0
  Else
    AData := StrToInt(locBuffer);
end;

procedure TXmlFormatter.GetStr(
  const ATypeInfo: PTypeInfo;
  var   AName: String;
  var   AData: String
);
begin
  AData := GetNodeValue(AName);
end;

procedure TXmlFormatter.GetObj(
  const ATypeInfo : PTypeInfo;
  var   AName     : String;
  var   AData     : TObject
);
Var
  propList : PPropList;
  i, propCount : Integer;
  pt : PTypeInfo;
  propName : String;
  intData : Integer;
  strData : String;
  objData : TObject;
  boolData : Boolean;
  p : PPropInfo;
  localedObj : Boolean;
begin
  If Assigned(AData) Then
    localedObj := False
  Else Begin
    AData := TBaseRemotableClass(GetTypeData(ATypeInfo)^.ClassType).Create();
    localedObj := True;
  End;
  Try
    TBaseRemotableClass(GetTypeData(ATypeInfo)^.ClassType).Load(AData as TBaseRemotable, Self,AName,ATypeInfo);
  Except
    If localedObj Then
      FreeAndNil(AData);
    Raise;
  End;
end;

function TXmlFormatter.GetCurrentScope(): String;
begin
  CheckScope();
  Result:= GetCurrentScopeObject().NodeName;
end;

function TXmlFormatter.GetCurrentScopeObject(): TDOMElement;
begin
  CheckScope();
  Result := FStack.Peek() As TDOMElement;
end;

constructor TXmlFormatter.Create();
begin
  FStack := TObjectStack.Create();
  FDoc := TXMLDocument.Create();
  FDoc.Encoding := 'UTF-8';
end;

destructor TXmlFormatter.Destroy();
begin
  FDoc.Free();
  FStack.Free();
  inherited Destroy();
end;

procedure TXmlFormatter.Clear();
begin
  ClearStack();
  FreeAndNil(FDoc);
  FDoc := TXMLDocument.Create();
  inherited Clear();
end;

procedure TXmlFormatter.BeginScope(
  Const AScopeName,ANameSpace : string;
  Const ANameSpaceShortName   : string
);
Var
  nsStr, scpStr : String;
  e : TDOMElement;
  hasNmspc, addAtt : Boolean;
begin
  scpStr := AScopeName;
  hasNmspc := Not IsStrEmpty(ANameSpace);
  If hasNmspc Then Begin
    nsStr := FindAttributeByValueInScope(ANameSpace);
    addAtt := IsStrEmpty(nsStr);
    If addAtt Then Begin
      If IsStrEmpty(ANameSpaceShortName) Then
        nsStr := 'ns' + IntToStr(NextNameSpaceCounter())
      Else
        nsStr := Trim(ANameSpaceShortName);
    End Else Begin
      nsStr := Copy(nsStr,Succ(AnsiPos(':',nsStr)),MaxInt);
    End;
    scpStr := nsStr + ':' + scpStr;
  End;

  e := FDoc.CreateElement(scpStr);
  If HasScope() Then
    GetCurrentScopeObject().AppendChild(e)
  Else
    FDoc.AppendChild(e);
  FStack.Push(e);
  If hasNmspc And addAtt Then
    e.SetAttribute('xmlns:'+nsStr,ANameSpace);
end;

procedure TXmlFormatter.EndScope();
begin
  CheckScope();
  FStack.Pop();
end;

procedure TXmlFormatter.AddScopeAttribute(const AName, AValue: string);
begin
  CheckScope();
  GetCurrentScopeObject().SetAttribute(AName,AValue);
end;

procedure TXmlFormatter.Put(
  const AName: String;
  const ATypeInfo: PTypeInfo;
  const AData
);
Var
  intData : Integer;
  strData : string;
  objData : TObject;
  boolData : Boolean;
begin
  Case ATypeInfo^.Kind Of
    tkInteger,tkInt64,tkQWord :
      Begin
        intData := Integer(AData);
        PutInt(AName,ATypeInfo,intData);
      End;
    tkLString, tkAString :
      Begin
        strData := String(AData);
        PutStr(AName,ATypeInfo,strData);
      End;
    tkClass :
      Begin
        objData := TObject(AData);
        PutObj(AName,ATypeInfo,objData);
      End;
    tkBool :
      Begin
        boolData := Boolean(AData);
        PutBool(AName,ATypeInfo,boolData);
      End;
  End;
end;

procedure TXmlFormatter.Get(
  const ATypeInfo : PTypeInfo;
  var   AName     : String;
  var   AData
);
Var
  intData : Integer;
  strData : string;
  objData : TObject;
  boolData : Boolean;
begin
  Case ATypeInfo^.Kind Of
    tkInteger,tkInt64,tkQWord :
      Begin
        intData := 0;
        GetInt(ATypeInfo,AName,intData);
        Integer(AData) := intData;
      End;
    tkLString, tkAString :
      Begin
        strData := '';
        GetStr(ATypeInfo,AName,strData);
        String(AData) := strData;
      End;
    tkClass :
      Begin
        objData := TObject(AData);
        GetObj(ATypeInfo,AName,objData);
        TObject(AData) := objData;
      End;
    tkBool :
      Begin
        boolData := False;
        GetBool(ATypeInfo,AName,boolData);
        Boolean(AData) := boolData;
      End;
  End;
end;

procedure TXmlFormatter.BeginCall(Const AProcName,ATarget,ACallAdress:string);
begin
  Inherited BeginCall(AProcName,ATarget,ACallAdress);
  BeginScope('Envelope',sSOAP_ENV,'SOAP-ENV');
    AddScopeAttribute('xmlns:xsi',sXSI_NS);
    AddScopeAttribute('xmlns:'+sXSD, sXSD_NS);
    BeginScope('Body',sSOAP_ENV);
        BeginScope(AProcName,ATarget);
end;

procedure TXmlFormatter.EndCall();
begin
      EndScope(); //BeginScope(AProcName,ATarget);
    EndScope();   //BeginScope('Body','http://schemas.xmlsoap.org/soap/envelope/');
  EndScope();     //BeginScope('Envelope','http://schemas.xmlsoap.org/soap/envelope/','SOAP-ENV');
end;

procedure TXmlFormatter.BeginCallRead();
Var
  nd : TDOMNode;
begin
  ClearStack();
  // DOM has to handle namespaces to avoid this !!!!
  nd := FDoc.DocumentElement.FindNode(sSOAP_ENV_ABR+':Body');
  Assert(Assigned(nd));
  FStack.Push(nd);
  nd := nd.FirstChild;
  Assert(Assigned(nd));
  FStack.Push(nd);
end;

procedure TXmlFormatter.SaveToStream(AStream: TStream);
begin
  WriteXMLFile(FDoc,AStream);
end;

procedure TXmlFormatter.LoadFromStream(AStream: TStream);
begin
  Clear();
  ReadXMLFile(FDoc,AStream);
end;

{ TSOAPRemotableComplexObject }

class procedure TSOAPRemotableComplexObject.Save(
        AObject   : TBaseRemotable;
        AStore    : TAbstractFormatter;
  const AName     : String;
  const ATypeInfo : PTypeInfo
);
Var
  typData : TTypeRegistryItem;
  nmspc : String;
  propList : PPropList;
  i, propCount : Integer;
  pt : PTypeInfo;
  intData : Integer;
  strData : String;
  objData : TObject;
  boolData : Boolean;
  p : PPropInfo;
begin
  typData := GetTypeRegistry().Find(ATypeInfo,False);
  If Assigned(typData) Then
    nmspc := typData.NameSpace
  Else
    nmspc := '';
  AStore.BeginScope(AName,'');
  Try
    AStore.AddScopeAttribute(sXSI_TYPE,nmspc);
    propCount := GetTypeData(ATypeInfo)^.PropCount;
    If ( propCount > 0 ) Then Begin
      GetPropList(ATypeInfo,propList);
      For i := 0 To Pred(propCount) Do Begin
        p := propList^[i];
        pt := p^.PropType;
        Case pt^.Kind Of
          tkInteger,tkInt64,tkQWord :
            Begin
              intData := GetOrdProp(AObject,p^.Name);
              AStore.Put(p^.Name,pt,intData);
            End;
          tkLString, tkAString :
            Begin
              strData := GetStrProp(AObject,p^.Name);
              AStore.Put(p^.Name,pt,strData);
            End;
          tkClass :
            Begin
              objData := GetObjectProp(AObject,p^.Name);
              AStore.Put(p^.Name,pt,objData);
            End;
          tkBool :
            Begin
              boolData := Boolean(GetOrdProp(AObject,p^.Name));
              AStore.Put(p^.Name,pt,boolData);
            End;
        End;
      End;
    End;
  Finally
    AStore.EndScope();
  End;
end;

class procedure TSOAPRemotableComplexObject.Load(
         AObject  : TBaseRemotable;
         AStore   : TAbstractFormatter;
  var   AName     : String;
  const ATypeInfo : PTypeInfo
);
Var
  propList : PPropList;
  i, propCount : Integer;
  pt : PTypeInfo;
  propName : String;
  intData : Integer;
  strData : String;
  objData : TObject;
  boolData : Boolean;
  p : PPropInfo;
begin
  AStore.BeginScopeRead(AName);
  Try
    propCount := GetTypeData(ATypeInfo)^.PropCount;
    If ( propCount > 0 ) Then Begin
      GetPropList(ATypeInfo,propList);
      For i := 0 To Pred(propCount) Do Begin
        p := propList^[i];
        If Assigned(p^.SetProc) Then Begin
          pt := p^.PropType;
          propName := p^.Name;
          Case pt^.Kind Of
            tkInteger,tkInt64,tkQWord :
              Begin
                AStore.Get(pt,propName,intData);
                SetOrdProp(AObject,propName,intData);
              End;
            tkLString, tkAString :
              Begin
                AStore.Get(pt,propName,strData);
                SetStrProp(AObject,propName,strData);
              End;
            tkBool :
              Begin
                AStore.Get(pt,propName,boolData);
                SetOrdProp(AObject,propName,Ord(boolData));
              End;
            tkClass :
              Begin
                objData := GetObjectProp(AObject,propName);
                Assert(Assigned(objData));
                Try
                  AStore.Get(pt,propName,objData);
                Except
                  FreeAndNil(objData);
                  Raise;
                End;
              End;
          End;
        End;
      End;
    End;
  Finally
    AStore.EndScopeRead();
  End;
end;

end.
